use std::sync::Arc;

use bevy_ecs::prelude::*;
use glam::{Mat4, Quat, Vec3};
use vulkano::{
    Validated,
    buffer::{AllocateBufferError, BufferContents, BufferCreateInfo, BufferUsage, Subbuffer},
    memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator},
    pipeline::graphics::vertex_input::Vertex,
};

use crate::core::render::primitives::{AsBindableBuffer, AsVertexBuffer};

use super::command::AsRenderableMeshInstance;

#[derive(Component, Debug, Clone)]
pub struct Transform {
    pub position: Vec3,
    pub rotation: Quat,
    pub scale: Vec3,
}

#[derive(BufferContents, Vertex, Clone, Copy, Debug)]
#[repr(C)]
pub struct TransformRaw {
    #[format(R32G32B32A32_SFLOAT)]
    pub model: [[f32; 4]; 4],
}

impl Default for Transform {
    fn default() -> Self {
        Self {
            position: Vec3::ZERO,
            rotation: Quat::IDENTITY,
            scale: Vec3::ONE,
        }
    }
}

impl Transform {
    pub fn new(position: Vec3, rotation: Quat, scale: Vec3) -> Self {
        Self {
            position,
            rotation,
            scale,
        }
    }

    pub fn rotate(&mut self, rotation: Quat) {
        self.rotation *= rotation;
    }

    pub fn translate(&mut self, translation: Vec3) {
        self.position += translation;
    }

    pub fn scale(&mut self, scale: Vec3) {
        self.scale *= scale;
    }

    pub fn set_position(&mut self, position: Vec3) {
        self.position = position;
    }

    pub fn set_rotation(&mut self, rotation: Quat) {
        self.rotation = rotation;
    }

    pub fn set_scale(&mut self, scale: Vec3) {
        self.scale = scale;
    }

    /// Get the transformation matrix (immutable - recalculates each time)
    pub fn matrix(&self) -> Mat4 {
        Mat4::from_translation(self.position)
            * Mat4::from_quat(self.rotation)
            * Mat4::from_scale(self.scale)
    }

    /// Convert to GPU-ready format (immutable - recalculates each time)
    pub fn to_raw(&self) -> TransformRaw {
        TransformRaw {
            model: self.matrix().to_cols_array_2d(),
        }
    }

    /// Create a buffer from transforms (immutable - recalculates each time)
    pub fn create_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        transforms: &[Transform],
    ) -> Result<Subbuffer<[TransformRaw]>, Validated<AllocateBufferError>> {
        TransformRaw::create_vertex_buffer(
            memory_allocator,
            &transforms.iter().map(|t| t.to_raw()).collect::<Vec<_>>(),
        )
    }
}

impl From<&Transform> for TransformRaw {
    fn from(transform: &Transform) -> Self {
        transform.to_raw()
    }
}

impl AsBindableBuffer for TransformRaw {
    type BufferData = TransformRaw;

    fn buffer_create_info() -> BufferCreateInfo {
        BufferCreateInfo {
            usage: BufferUsage::VERTEX_BUFFER,
            ..Default::default()
        }
    }

    fn allocation_create_info() -> AllocationCreateInfo {
        AllocationCreateInfo {
            memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
            ..Default::default()
        }
    }

    fn to_buffer_data(&self) -> Self::BufferData {
        *self
    }
}

impl AsVertexBuffer for TransformRaw {}

impl AsRenderableMeshInstance for Subbuffer<[TransformRaw]> {
    type InstanceBufferData = TransformRaw;

    fn instance_buffer(&self) -> &Subbuffer<[Self::InstanceBufferData]> {
        self
    }

    fn instance_count(&self) -> u32 {
        self.len() as u32
    }
}
