use std::{collections::HashMap, error::Error, sync::Arc};

use vulkano::{
    command_buffer::{
        AutoCommandBufferBuilder, CommandBufferUsage, PrimaryCommandBufferAbstract,
        allocator::StandardCommandBufferAllocator,
    },
    device::{Device, Queue},
    format::Format,
    image::sampler::SamplerCreateInfo,
    memory::allocator::StandardMemoryAllocator,
};

use crate::core::render::primitives::vulkan_resource::{
    VulkanCommandBufferAllocator, VulkanDevice, VulkanGraphicsQueue, VulkanMemoryAllocator,
    VulkanTransferQueue,
};
use bevy_ecs::{resource::Resource, world::World};

use super::Texture;

pub enum TextureSourceKind {
    File(String),
    Buffer(Vec<u8>),
}

pub struct TextureLoadInfo {
    pub source: TextureSourceKind,
    pub sampler_create_info: SamplerCreateInfo,
    pub image_format: Format,
}

#[derive(Resource)]
pub struct TextureLoader {
    loaded_textures: HashMap<String, Arc<Texture>>,
    pending_textures: HashMap<String, TextureLoadInfo>,
    device: Arc<Device>,
    command_buffer_allocator: Arc<StandardCommandBufferAllocator>,
    memory_allocator: Arc<StandardMemoryAllocator>,
    queue: Arc<Queue>,
}

impl TextureLoader {
    pub fn new(
        device: Arc<Device>,
        command_buffer_allocator: Arc<StandardCommandBufferAllocator>,
        memory_allocator: Arc<StandardMemoryAllocator>,
        queue: Arc<Queue>,
    ) -> Self {
        Self {
            loaded_textures: HashMap::new(),
            pending_textures: HashMap::new(),
            device,
            command_buffer_allocator,
            memory_allocator,
            queue,
        }
    }

    pub fn select_best_suitable_queue(
        graphics_queue: &Arc<Queue>,
        transfer_queue: Option<&Arc<Queue>>,
    ) -> Arc<Queue> {
        transfer_queue
            .map(|queue| {
                tracing::trace!(
                    "Select transfer queue for texture loading with family index: {:?}",
                    queue.queue_family_index()
                );
                queue.clone()
            })
            .unwrap_or_else(|| {
                tracing::trace!(
                    "Select graphics queue for texture loading with family index: {:?}",
                    graphics_queue.queue_family_index()
                );
                graphics_queue.clone()
            })
    }

    pub fn add_texture(&mut self, name: String, load_info: TextureLoadInfo) {
        self.pending_textures.insert(name, load_info);
    }

    pub fn load_pending_textures(&mut self) -> Result<(), Box<dyn Error>> {
        let _span = tracing::info_span!("load_pending_textures");
        let mut uploads = AutoCommandBufferBuilder::primary(
            self.command_buffer_allocator.clone(),
            self.queue.queue_family_index(),
            CommandBufferUsage::OneTimeSubmit,
        )?;

        let mut loading_textures = HashMap::new();

        tracing::trace!("Pending textures count: {}", self.pending_textures.len());

        for (name, info) in self.pending_textures.iter() {
            let texture = match &info.source {
                TextureSourceKind::File(path) => Texture::from_file(
                    &self.device,
                    &self.memory_allocator,
                    &mut uploads,
                    path.as_str(),
                    info,
                )?,
                TextureSourceKind::Buffer(buffer) => Texture::from_bytes(
                    &self.device,
                    &self.memory_allocator,
                    &mut uploads,
                    &buffer,
                    info,
                )?,
            };

            loading_textures.insert(name.clone(), Arc::new(texture));
            tracing::trace!("Loaded texture: {}", name);
        }

        let _ = uploads.build()?.execute(self.queue.clone())?;

        self.loaded_textures.extend(loading_textures);

        Ok(())
    }

    pub fn get_texture(&self, name: &str) -> Option<&Arc<Texture>> {
        self.loaded_textures.get(name)
    }

    pub fn pending_textures_count(&self) -> usize {
        self.pending_textures.len()
    }

    pub fn loaded_textures_count(&self) -> usize {
        self.loaded_textures.len()
    }
}
